# -*- coding: utf-8 -*-
# Copyright (c) 2018 the Pockets team, see AUTHORS.
# Licensed under the BSD License, see LICENSE for details.

"""A pocket full of useful decorators!"""

from __future__ import absolute_import, print_function

import inspect
from functools import wraps

from pockets.collections import listify
from pockets.inspect import unwrap


__all__ = [
    "argmod",
    "cached_classproperty",
    "cached_property",
    "classproperty",
]


def argmod(*args):
    """
    Decorator that intercepts and modifies function arguments.

    Args:
        from_param (str|list): A parameter or list of possible parameters that
            should be modified using `modifier_func`. Passing a list of
            possible parameters is useful when a function's parameter names
            have changed, but you still want to support the old parameter
            names.
        to_param (str): Optional. If given, to_param will be used as the
            parameter name for the modified argument. If not given, to_param
            will default to the last parameter given in `from_param`.
        modifier_func (callable): The function used to modify the `from_param`.

    Returns:
        function: A function that modifies the given `from_param` before the
            function is called.
    """
    from_param = listify(args[0])
    to_param = from_param[-1] if len(args) < 3 else args[1]
    modifier_func = args[-1]

    def _decorator(func):
        try:
            argspec = inspect.getfullargspec(unwrap(func))
        except AttributeError:
            argspec = inspect.getargspec(unwrap(func))
        if to_param not in argspec.args:
            return func
        arg_index = argspec.args.index(to_param)

        @wraps(func)
        def _modifier(*args, **kwargs):
            kwarg = False
            for arg in from_param:
                if arg in kwargs:
                    kwarg = arg
                    break

            if kwarg:
                kwargs[to_param] = modifier_func(kwargs.pop(kwarg))
            elif arg_index < len(args):
                args = list(args)
                args[arg_index] = modifier_func(args[arg_index])
            return func(*args, **kwargs)

        return _modifier

    return _decorator


class cached_classproperty(property):
    """
    Like @cached_property except it works on classes instead of instances.


    Note:
        Class properties created by @cached_classproperty are read-only.
        Any attempts to write to the property will erase the
        @cached_classproperty, and the behavior of the underlying method
        will be lost.

    >>> class MyClass(object):
    ...     @cached_classproperty
    ...     def myproperty(cls):
    ...         return '{0}.myproperty'.format(cls.__name__)
    >>> MyClass.myproperty
    'MyClass.myproperty'

    """

    def __init__(self, fget, *arg, **kw):
        super(cached_classproperty, self).__init__(fget, *arg, **kw)
        self.__doc__ = fget.__doc__
        self.__fget_name__ = fget.__name__

    def __get__(desc, self, cls):
        cache_attr = "_cached_{0}_{1}".format(cls.__name__, desc.__fget_name__)
        if not hasattr(cls, cache_attr):
            setattr(cls, cache_attr, desc.fget(cls))
        return getattr(cls, cache_attr)

    def getter(self, fget):
        raise AttributeError("@cached_classproperty.getter is not supported")

    def setter(self, fset):
        raise AttributeError("@cached_classproperty.setter is not supported")

    def deleter(self, fdel):
        raise AttributeError("@cached_classproperty.deleter is not supported")


def cached_property(func):
    """Decorator for making readonly, memoized properties."""
    cache_attr = "_cached_{0}".format(func.__name__)

    @property
    @wraps(func)
    def caching(self, *args, **kwargs):
        if not hasattr(self, cache_attr):
            setattr(self, cache_attr, func(self, *args, **kwargs))
        return getattr(self, cache_attr)

    return caching


class classproperty(property):
    """
    Decorator to create a read-only class property similar to classmethod.

    For whatever reason, the @property decorator isn't smart enough to
    recognize @classmethods and behaves differently on them than on instance
    methods.  This decorator may be used like to create a class-level property,
    useful for singletons and other one-per-class properties.

    This implementation is partially based on
    `sqlalchemy.util.langhelpers.classproperty`.

    Note:
        Class properties created by @classproperty are read-only. Any attempts
        to write to the property will erase the @classproperty, and the
        behavior of the underlying method will be lost.

    >>> class MyClass(object):
    ...     @classproperty
    ...     def myproperty(cls):
    ...         return '{0}.myproperty'.format(cls.__name__)
    >>> MyClass.myproperty
    'MyClass.myproperty'

    """

    def __init__(self, fget, *arg, **kw):
        super(classproperty, self).__init__(fget, *arg, **kw)
        self.__doc__ = fget.__doc__

    def __get__(desc, self, cls):
        return desc.fget(cls)

    def getter(self, fget):
        raise AttributeError("@classproperty.getter is not supported")

    def setter(self, fset):
        raise AttributeError("@classproperty.setter is not supported")

    def deleter(self, fdel):
        raise AttributeError("@classproperty.deleter is not supported")



# -*- coding: utf-8 -*-
# Copyright (c) 2018 the Pockets team, see AUTHORS.
# Licensed under the BSD License, see LICENSE for details.

"""A pocket full of useful reflection functions!"""

from __future__ import absolute_import, print_function

import inspect
import functools
from os.path import basename
from pkgutil import iter_modules

import six
from six import string_types

from pockets.collections import listify
from pockets.string import splitify


__all__ = [
    "collect_subclasses",
    "collect_superclasses",
    "collect_superclass_attr_names",
    "hoist_submodules",
    "import_star",
    "import_submodules",
    "is_data",
    "resolve",
    "unwrap",
]


def collect_subclasses(cls):
    """
    Recursively collects all descendant subclasses that inherit from the
    given class, not including the class itself.

    Note:
        Does not include `cls` itself.

    Args:
        cls (class): The class object from which the collection should begin.

    Returns:
        list: A list of `class` objects that inherit from `cls`. This list
            will not include `cls` itself.
    """
    subclasses = set()
    for subclass in cls.__subclasses__():
        subclasses.add(subclass)
        subclasses.update(collect_subclasses(subclass))
    return list(subclasses)


def collect_superclasses(cls, terminal_class=None, modules=None):
    """
    Recursively collects all ancestor superclasses in the inheritance
    hierarchy of the given class, including the class itself.

    Note:
        Inlcudes `cls` itself. Will not include `terminal_class`.

    Args:
        cls (class): The class object from which the collection should begin.
        terminal_class (class or list): If `terminal_class` is encountered in
            the hierarchy, we stop ascending the tree. `terminal_class` will
            not be included in the returned list.
        modules (string, module, or list): If `modules` is passed, we only
            return classes that are in the given module/modules. This can be
            used to exclude base classes that come from external libraries.

    Returns:
        list: A list of `class` objects from which `cls` inherits. This list
            will include `cls` itself.
    """
    terminal_class = listify(terminal_class)
    if modules is not None:
        modules = listify(modules)
        module_strings = []
        for m in modules:
            if isinstance(m, six.string_types):
                module_strings.append(m)
            else:
                module_strings.append(m.__name__)
        modules = module_strings

    superclasses = set()
    is_in_module = modules is None or cls.__module__ in modules
    if is_in_module and cls not in terminal_class:
        superclasses.add(cls)
        for base in cls.__bases__:
            superclasses.update(
                collect_superclasses(base, terminal_class, modules)
            )

    return list(superclasses)


def collect_superclass_attr_names(cls, terminal_class=None, modules=None):
    """
    Recursively collects all attribute names of ancestor superclasses in the
    inheritance hierarchy of the given class, including the class itself.

    Note:
        Inlcudes `cls` itself. Will not include `terminal_class`.

    Args:
        cls (class): The class object from which the collection should begin.
        terminal_class (class or list): If `terminal_class` is encountered in
            the hierarchy, we stop ascending the tree. Attributes from
            `terminal_class` will not be included in the returned list.
        modules (string, module, or list): If `modules` is passed, we only
            return classes that are in the given module/modules. This can be
            used to exclude base classes that come from external libraries.

    Returns:
        list: A list of `str` attribute names for every `class` in the
            inheritance hierarchy.
    """
    superclasses = collect_superclasses(cls, terminal_class, modules)
    attr_names = set()
    for superclass in superclasses:
        attr_names.update(superclass.__dict__.keys())
    return list(attr_names)


def hoist_submodules(package, extend_all=True):
    """
    Sets `__all__` attrs from submodules of `package` as attrs on `package`.

    Note:
        This only considers attributes exported by `__all__`. If a submodule
        does not define `__all__`, then it is ignored.

    Effectively does::

        from package.* import *

    Args:
        package (str or module): The parent package into which submodule
            exports should be hoisted.
        extend_all (bool): If True, `package.__all__` will be extended
            to include the hoisted attributes. Defaults to True.

    Returns:
        list: List of all hoisted attribute names.

    """
    module = resolve(package)
    hoisted_attrs = []
    for submodule in import_submodules(module):
        for attr_name, attr in import_star(submodule).items():
            hoisted_attrs.append(attr_name)
            setattr(module, attr_name, attr)

    if extend_all:
        if getattr(module, "__all__", None) is None:
            module.__all__ = list(hoisted_attrs)
        else:
            module.__all__.extend(hoisted_attrs)

    return hoisted_attrs


def import_star(module):
    """
    Imports all exported attributes of `module` and returns them in a `dict`.

    Note:
        This only considers attributes exported by `__all__`. If `module`
        does not define `__all__`, then nothing is imported.

    Effectively does::

        from module import *

    Args:
        module (str or module): The module from which a wildcard import
            should be done.

    Returns:
        dict: Map of all imported attributes.

    """
    module = resolve(module)
    attrs = getattr(module, "__all__", [])
    return dict([(attr, getattr(module, attr)) for attr in attrs])


def import_submodules(package):
    """
    Imports all submodules of `package`.

    Effectively does::

        __import__(package.*)

    Args:
        package (str or module): The parent package from which submodules
            should be imported.

    Yields:
        module: The next submodule of `package`.

    """
    module = resolve(package)
    if basename(module.__file__).startswith("__init__.py"):
        for _, submodule_name, _ in iter_modules(module.__path__):
            yield resolve(submodule_name, module)


def is_data(obj):
    """
    Returns True if `obj` is a "data like" object.

    Strongly inspired by `inspect.classify_class_attrs`. This function is
    useful when trying to determine if an attribute has a meaningful docstring
    or not. In general, a routine can have meaningful docstrings, whereas
    non-routines cannot.

    See Also:
        * `inspect.classify_class_attrs`
        * `inspect.isroutine`

    Args:
        obj (object): The object in question.

    Returns:
        bool: True if `obj` is "data like", False otherwise.
    """
    if isinstance(
        obj, (staticmethod, classmethod, property)
    ) or inspect.isroutine(obj):
        return False
    else:
        return True


def resolve(name, modules=None):
    """
    Resolve a dotted name to an object (usually class, module, or function).

    If `name` is a string, attempt to resolve it according to Python
    dot notation, e.g. "path.to.MyClass". If `name` is anything other than a
    string, return it immediately:

    >>> resolve("calendar.TextCalendar")
    <class 'calendar.TextCalendar'>
    >>> resolve(object())
    <object object at 0x...>

    If `modules` is specified, then resolution of `name` is restricted
    to the given modules. Leading dots are allowed in `name`, but they are
    ignored. Resolution **will not** traverse up the module path if `modules`
    is specified.

    If `modules` is not specified and `name` has leading dots, then resolution
    is first attempted relative to the calling function's module, and then
    absolutely. Resolution **will** traverse up the module path. If `name` has
    no leading dots, resolution is first attempted absolutely and then
    relative to the calling module.

    Pass an empty string for `modules` to only use absolute resolution.

    Warning:
        Do not resolve strings supplied by an end user without specifying
        `modules`. Instantiating an arbitrary object specified by an end user
        can introduce a potential security risk.

        To avoid this, restrict the search path by explicitly specifying
        `modules`.

    Restricting `name` resolution to a set of `modules`:

    >>> resolve("pockets.camel")
    <function camel at 0x...>
    >>> resolve("pockets.camel", modules=["re", "six"])
    Traceback (most recent call last):
    ValueError: Unable to resolve 'pockets.camel' in modules: ['re', 'six']
      ...

    Args:
        name (str or object): A dotted name.

        modules (str, module, or list, optional): A module or list of modules,
            under which to search for `name`.

    Returns:
        object: The object specified by `name`.

    Raises:
        ValueError: If `name` can't be resolved.

    """
    if not isinstance(name, string_types):
        return name

    obj_path = splitify(name, ".", include_empty=True)
    search_paths = []
    if modules is not None:
        while not obj_path[0].strip():
            obj_path.pop(0)
        for module_path in listify(modules):
            search_paths.append(splitify(module_path, ".") + obj_path)
    else:
        caller = inspect.getouterframes(inspect.currentframe())[1][0].f_globals
        module_path = caller["__name__"].split(".")
        if not obj_path[0]:
            obj_path.pop(0)
            while not obj_path[0]:
                obj_path.pop(0)
                if module_path:
                    module_path.pop()

            search_paths.append(module_path + obj_path)
            search_paths.append(obj_path)
        else:
            search_paths.append(obj_path)
            search_paths.append(module_path + obj_path)

    exceptions = []
    for path in search_paths:
        # Import the most deeply nested module available
        module = None
        module_path = []
        obj_path = list(path)
        while obj_path:
            module_name = obj_path.pop(0)
            while not module_name:
                module_name = obj_path.pop(0)
            if isinstance(module_name, string_types):
                package = ".".join(module_path + [module_name])
                try:
                    module = __import__(package, fromlist=module_name)
                except ImportError as ex:
                    exceptions.append(ex)
                    obj_path = [module_name] + obj_path
                    break
                else:
                    module_path.append(module_name)
            else:
                module = module_name
                module_path.append(module.__name__)

        if module:
            if obj_path:
                try:
                    return functools.reduce(getattr, obj_path, module)
                except AttributeError as ex:
                    exceptions.append(ex)
            else:
                return module

    if modules:
        msg = "Unable to resolve '{0}' in modules: {1}".format(name, modules)
    else:
        msg = "Unable to resolve '{0}'".format(name)

    if exceptions:
        msgs = ["{0}: {1}".format(type(e).__name__, e) for e in exceptions]
        raise ValueError("\n    ".join([msg] + msgs))
    else:
        raise ValueError(msg)


def unwrap(func):
    """
    Finds the innermost function that has been wrapped using `functools.wrap`.

    Note:
        This function relies on the existence of the `__wrapped__` attribute,
        which was not automatically added until Python 3.2. If you are using
        an older version of Python, you'll have to manually add the
        `__wrapped__` attribute in order to use `unwrap`::

            def my_decorator(func):
                @wraps(func)
                def with_my_decorator(*args, **kwargs):
                    return func(*args, **kwargs)

                if not hasattr(with_my_decorator, '__wrapped__'):
                    with_my_decorator.__wrapped__ = func

                return with_my_decorator

    Args:
        func (function): A function that may or may not have been wrapped
            using `functools.wrap`.

    Returns:
        function: The original function before it was wrapped using
            `functools.wrap`. `func` is returned directly, if it was never
            wrapped using `functools.wrap`.
    """
    return unwrap(func.__wrapped__) if hasattr(func, "__wrapped__") else func



# -*- coding: utf-8 -*-
# Copyright (c) 2018 the Pockets team, see AUTHORS.
# Licensed under the BSD License, see LICENSE for details.

"""A pocket full of useful collection tools!"""

from __future__ import absolute_import, print_function

from collections import defaultdict
from inspect import isclass

try:
    from collections.abc import Iterable, Mapping, Sized
except ImportError:
    from collections import Iterable, Mapping, Sized
try:
    from collections import OrderedDict
except ImportError:
    OrderedDict = dict

import six


__all__ = [
    "groupify",
    "keydefaultdict",
    "is_listy",
    "listify",
    "is_mappy",
    "mappify",
    "nesteddefaultdict",
    "readable_join",
    "uniquify",
]


def groupify(items, keys, val_key=None):
    """
    Groups a list of items into nested OrderedDicts based on the given keys.

    Note:
        On Python 2.6 the return value will use regular dicts instead of
        OrderedDicts.

    >>> from __future__ import print_function
    >>> from json import dumps
    >>>
    >>> ex = lambda x: print(dumps(x, indent=2, sort_keys=True, default=repr))
    >>>
    >>> class Reminder:
    ...   def __init__(self, when, where, what):
    ...     self.when = when
    ...     self.where = where
    ...     self.what = what
    ...   def __repr__(self):
    ...     return 'Reminder({0.when}, {0.where}, {0.what})'.format(self)
    ...
    >>> reminders = [
    ...   Reminder('Fri', 'Home', 'Eat cereal'),
    ...   Reminder('Fri', 'Work', 'Feed Ivan'),
    ...   Reminder('Sat', 'Home', 'Sleep in'),
    ...   Reminder('Sat', 'Home', 'Play Zelda'),
    ...   Reminder('Sun', 'Home', 'Sleep in'),
    ...   Reminder('Sun', 'Work', 'Reset database')]
    >>>
    >>> ex(groupify(reminders, 'when'))
    {
      "Fri": [
        "Reminder(Fri, Home, Eat cereal)",
        "Reminder(Fri, Work, Feed Ivan)"
      ],
      "Sat": [
        "Reminder(Sat, Home, Sleep in)",
        "Reminder(Sat, Home, Play Zelda)"
      ],
      "Sun": [
        "Reminder(Sun, Home, Sleep in)",
        "Reminder(Sun, Work, Reset database)"
      ]
    }
    >>>
    >>> ex(groupify(reminders, ['when', 'where']))
    {
      "Fri": {
        "Home": [
          "Reminder(Fri, Home, Eat cereal)"
        ],
        "Work": [
          "Reminder(Fri, Work, Feed Ivan)"
        ]
      },
      "Sat": {
        "Home": [
          "Reminder(Sat, Home, Sleep in)",
          "Reminder(Sat, Home, Play Zelda)"
        ]
      },
      "Sun": {
        "Home": [
          "Reminder(Sun, Home, Sleep in)"
        ],
        "Work": [
          "Reminder(Sun, Work, Reset database)"
        ]
      }
    }
    >>>
    >>> ex(groupify(reminders, ['when', 'where'], 'what'))
    {
      "Fri": {
        "Home": [
          "Eat cereal"
        ],
        "Work": [
          "Feed Ivan"
        ]
      },
      "Sat": {
        "Home": [
          "Sleep in",
          "Play Zelda"
        ]
      },
      "Sun": {
        "Home": [
          "Sleep in"
        ],
        "Work": [
          "Reset database"
        ]
      }
    }
    >>>
    >>> ex(groupify(reminders, lambda r: '{0.when} - {0.where}'.format(r), 'what'))
    {
      "Fri - Home": [
        "Eat cereal"
      ],
      "Fri - Work": [
        "Feed Ivan"
      ],
      "Sat - Home": [
        "Sleep in",
        "Play Zelda"
      ],
      "Sun - Home": [
        "Sleep in"
      ],
      "Sun - Work": [
        "Reset database"
      ]
    }

    Args:
        items (list): The list of items to arrange in groups.
        keys (str|callable|list): The key or keys that should be used to group
            `items`. If multiple keys are given, then each will correspond to
            an additional level of nesting in the order they are given.
        val_key (str|callable): A key or callable used to generate the leaf
            values in the nested OrderedDicts. If `val_key` is `None`, then
            the item itself is used. Defaults to `None`.

    Returns:
        OrderedDict: Nested OrderedDicts with `items` grouped by `keys`.

    """  # noqa: E501

    if not keys:
        return items
    keys = listify(keys)
    last_key = keys[-1]
    is_callable = callable(val_key)
    groupified = OrderedDict()
    for item in items:
        current = groupified
        for key in keys:
            attr = key(item) if callable(key) else getattr(item, key)
            if attr not in current:
                current[attr] = [] if key is last_key else OrderedDict()
            current = current[attr]
        if val_key:
            value = val_key(item) if is_callable else getattr(item, val_key)
        else:
            value = item
        current.append(value)
    return groupified


class keydefaultdict(defaultdict):
    """
    A defaultdict that passes the missed key to the factory function.

    >>> def echo_factory(missing_key):
    ...     return missing_key
    ...
    >>> d = keydefaultdict(echo_factory)
    >>> d['Hello World']
    'Hello World'
    >>> d['Hello World'] = 'Goodbye'
    >>> d['Hello World']
    'Goodbye'

    """

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        else:
            ret = self[key] = self.default_factory(key)
            return ret


def is_listy(x):
    """
    Return True if `x` is "listy", i.e. a list-like object.

    "Listy" is defined as a sized iterable which is neither a map nor a string:

    >>> is_listy(['a', 'b'])
    True
    >>> is_listy(set())
    True
    >>> is_listy(iter(['a', 'b']))
    False
    >>> is_listy({'a': 'b'})
    False
    >>> is_listy('a regular string')
    False

    Note:
        Iterables and generators fail the "listy" test because they
        are not sized.

    Args:
        x (any value): The object to test.

    Returns:
        bool: True if `x` is "listy", False otherwise.

    """
    return (
        isinstance(x, Sized)
        and isinstance(x, Iterable)
        and not isinstance(x, (Mapping, type(b"")))
        and not isinstance(x, six.string_types)
    )


def listify(x, minlen=0, default=None, cls=None):
    """
    Return a listified version of `x`.

    If `x` is a non-string iterable, it is wrapped in a list; otherwise
    a list is returned with `x` as its only element. If `x` is `None`, an
    empty list is returned.

    >>> listify('a regular string')
    ['a regular string']
    >>> listify(tuple(['a', 'b', 'c']))
    ['a', 'b', 'c']
    >>> listify({'a': 'A'})
    [{'a': 'A'}]
    >>> listify(None)
    []

    Note:
        Not guaranteed to return a copy of `x`. If `x` is already a list and
        `cls` is not specified, then `x` itself is returned.

    Args:
        x (any value): Value to listify.

        minlen (int): Minimum length of the returned list. If the returned
            list would be shorter than `minlen` it is padded with values from
            `default`. Defaults to 0.

            >>> listify([], minlen=0)
            []
            >>> listify([], minlen=1)
            [None]
            >>> listify('item', minlen=3)
            ['item', None, None]

        default (any value): Value that should be used to pad the list if it
            would be shorter than `minlen`:

            >>> listify([], minlen=1, default='PADDING')
            ['PADDING']
            >>> listify('item', minlen=3, default='PADDING')
            ['item', 'PADDING', 'PADDING']

        cls (class or callable): Instead of wrapping `x` in a list, wrap it
            in an instance of `cls`. `cls` should accept an iterable object
            as its single parameter when called:

            >>> from collections import deque
            >>> listify(['a', 'b', 'c'], cls=deque)
            deque(['a', 'b', 'c'])

    Returns:
        list or `cls`: A listified version of `x`.

    """
    if x is None:
        x = []
    elif not isinstance(x, list):
        x = list(x) if is_listy(x) else [x]

    if minlen and len(x) < minlen:
        x.extend([default for i in range(minlen - len(x))])

    if cls and not (isclass(cls) and issubclass(type(x), cls)):
        x = cls(x)
    return x


def is_mappy(x):
    """
    Return True if `x` is "mappy", i.e. a map-like object.

    "Mappy" is defined as any instance of `collections.Mapping`:

    >>> is_mappy({'a': 'b'})
    True
    >>> from collections import defaultdict
    >>> is_mappy(defaultdict(list))
    True
    >>> is_mappy('a regular string')
    False
    >>> is_mappy(['a', 'b'])
    False
    >>> is_listy(iter({'a': 'b'}))
    False

    Note:
        Iterables and generators fail the "mappy" test.

    Args:
        x (any value): The object to test.

    Returns:
        bool: True if `x` is "mappy", False otherwise.

    """
    return isinstance(x, Mapping)


def mappify(x, default=True, cls=None):
    """
    Return a mappified version of `x`.

    If `x` is a string, it becomes the only key of the returned dict. If `x`
    is a non-string iterable, the elements of `x` become keys in the returned
    dict. The values of the returned dict are set to `default`. If `x` is
    `None`, an empty dict is returned.

    If `x` is a map, it is returned directly.

    >>> mappify('a regular string')
    {'a regular string': True}
    >>> mappify(['a'])
    {'a': True}
    >>> mappify({'a': 'A'})
    {'a': 'A'}
    >>> mappify(None)
    {}

    Note:
        Not guaranteed to return a copy of `x`. If `x` is already a map and
        `cls` is not specified, then `x` itself is returned.

    Args:
        x (str, map, or iterable): Value to mappify.

        default (any value): Value used to fill out missing values of the
            returned dict.

        cls (class or callable): Instead of wrapping `x` in a dict, wrap it
            in an instance of `cls`. `cls` should accept a map object as
            its single parameter when called:

            >>> from collections import defaultdict
            >>> mappify('a', cls=lambda x: defaultdict(None, x))
            defaultdict(None, {'a': True})

    Returns:
        dict or `cls`: A mappified version of `x`.

    Raises:
        TypeError: If `x` is not a map, iterable, or string.

    """
    if x is None:
        x = {}
    elif not isinstance(x, Mapping):
        if isinstance(x, six.string_types):
            x = {x: default}
        elif isinstance(x, Iterable):
            # If cls is specified, attempt to preserve the order of x, in
            # case cls is also a class that preserves order.
            arg = [(v, default) for v in x]
            x = OrderedDict(arg) if cls else dict(arg)
        else:
            raise TypeError(
                "Unable to mappify non-mappy {0}".format(type(x)), x
            )

    if cls and not (isclass(cls) and issubclass(type(x), cls)):
        x = cls(x)
    return x


def nesteddefaultdict():
    """
    A defaultdict that returns nested defaultdicts as the default value.

    Each defaultdict returned as the default value will also return nested
    defaultdicts, and so on.

    >>> nested = nesteddefaultdict()
    >>> nested_child = nested['New Key 1']
    >>> nested_child
    defaultdict(...)
    >>> nested_grandchild = nested_child['New Key 2']
    >>> nested_grandchild
    defaultdict(...)

    """
    return defaultdict(nesteddefaultdict)


def readable_join(xs, conjunction="and", sep=","):
    """
    Accepts a list of strings and separates them with commas as grammatically
    appropriate with a conjunction before the final entry. Any input strings
    containing only whitespace will not be included in the result.

    >>> readable_join(['foo'])
    'foo'
    >>> readable_join(['foo', 'bar'])
    'foo and bar'
    >>> readable_join(['foo', 'bar', 'baz'])
    'foo, bar, and baz'
    >>> readable_join(['foo', '  ', '', 'bar', '', '  ', 'baz'])
    'foo, bar, and baz'
    >>> readable_join(['foo', 'bar', 'baz'], 'or')
    'foo, bar, or baz'
    >>> readable_join(['foo', 'bar', 'baz'], 'but never')
    'foo, bar, but never baz'

    """
    xs = [s for s in map(lambda s: str(s).strip(), listify(xs)) if s]
    if len(xs) > 1:
        xs = list(xs)
        xs[-1] = conjunction + " " + xs[-1]
    return (sep + " " if len(xs) > 2 else " ").join(xs)


def uniquify(x, key=lambda o: o, cls=None):
    """
    Returns an order-preserved copy of `x` with duplicate items removed.

    >>> uniquify(['a', 'z', 'a', 'b', 'a', 'y', 'a', 'c', 'a', 'x'])
    ['a', 'z', 'b', 'y', 'c', 'x']

    Args:
        x (Sequence): Sequence to uniquify.

        key (str or callable): Similar to `sorted`, specifies an attribute or
            function of one argument that is used to extract a comparison key
            from each list element: key=str.lower. By default, compares the
            elements directly.

            >>> strings = ['ASDF', 'asdf', 'ZXCV', 'zxcv']
            >>> uniquify(strings, key=str.lower)
            ['ASDF', 'ZXCV']

        cls (class or callable): Instead of wrapping `x` in a list, wrap it
            in an instance of `cls`. `cls` should accept an iterable object
            as its single parameter when called:

            >>> from collections import deque
            >>> listify(['a', 'b', 'c'], cls=deque)
            deque(['a', 'b', 'c'])

    Returns:
        list: An order-preserved copy of `x` with duplicate items removed.

    Raises:
        TypeError: If `x` is not "listy".

    """
    if not is_listy(x):
        raise TypeError("Unable to uniquify non-listy {0}".format(type(x)), x)
    seen = set()
    keys = [(key(o) if callable(key) else getattr(o, key), o) for o in x]
    x = [o for k, o in keys if k not in seen and not seen.add(k)]

    if cls and not (isclass(cls) and issubclass(type(x), cls)):
        x = cls(x)
    return x



# -*- coding: utf-8 -*-
# Copyright (c) 2018 the Pockets team, see AUTHORS.
# Licensed under the BSD License, see LICENSE for details.

"""A pocket full of useful iterators!"""

from __future__ import absolute_import, print_function

import collections

import six


__all__ = ["itermod", "iterpeek", "modify_iter", "peek_iter"]


class iterpeek(object):
    """
    An iterator object that supports peeking ahead.

    >>> p = iterpeek(["a", "b", "c", "d", "e"])
    >>> p.peek()
    'a'
    >>> p.next()
    'a'
    >>> p.peek(3)
    ['b', 'c', 'd']

    Args:
        o (iterable or callable): `o` is interpreted very differently
            depending on the presence of `sentinel`.

            If `sentinel` is not given, then `o` must be a collection object
            which supports either the iteration protocol or the sequence
            protocol.

            If `sentinel` is given, then `o` must be a callable object.

        sentinel (any value, optional): If given, the iterator will call `o`
            with no arguments for each call to its `next` method; if the value
            returned is equal to `sentinel`, :exc:`StopIteration` will be
            raised, otherwise the value will be returned.

    See Also:
        `iterpeek` can operate as a drop in replacement for the built-in
        `iter <https://docs.python.org/3/library/functions.html#iter>`_
        function.

    Attributes:
        sentinel (any value): The value used to indicate the iterator is
            exhausted. If `sentinel` was not given when the `iterpeek` was
            instantiated, then it will be set to a new object
            instance: ``object()``.

    """

    def __init__(self, *args):
        """__init__(o, sentinel=None)"""
        self._iterable = iter(*args)
        self._cache = collections.deque()
        self.sentinel = args[1] if len(args) > 1 else object()

    def __iter__(self):
        return self

    def __next__(self, n=None):
        # NOTE: Prevent 2to3 from transforming self.next() in next(self),
        # which causes an infinite loop!
        return getattr(self, "next")(n)

    def _fillcache(self, n):
        """Cache `n` items. If `n` is 0 or None, then 1 item is cached."""
        if not n:
            n = 1
        try:
            while len(self._cache) < n:
                self._cache.append(next(self._iterable))
        except StopIteration:
            while len(self._cache) < n:
                self._cache.append(self.sentinel)

    def has_next(self):
        """
        Determine if iterator is exhausted.

        Returns:
            bool: True if iterator has more items, False otherwise.

        Note:
            Will never raise :exc:`StopIteration`.

        """
        return self.peek() != self.sentinel

    def next(self, n=None):
        """
        Get the next item or `n` items of the iterator.

        Args:
            n (int, optional): The number of items to retrieve. Defaults to
                None.

        Returns:
            item or list of items: The next item or `n` items of the iterator.
            If `n` is None, the item itself is returned. If `n` is an int,
            the items will be returned in a list. If `n` is 0, an empty
            list is returned:

                >>> p = iterpeek(["a", "b", "c", "d", "e"])
                >>> p.next()
                'a'
                >>> p.next(0)
                []
                >>> p.next(1)
                ['b']
                >>> p.next(2)
                ['c', 'd']

        Raises:
            StopIteration: Raised if the iterator is exhausted, even if
                `n` is 0.

        """
        self._fillcache(n)
        if not n:
            if self._cache[0] == self.sentinel:
                raise StopIteration
            if n is None:
                result = self._cache.popleft()
            else:
                result = []
        else:
            if self._cache[n - 1] == self.sentinel:
                raise StopIteration
            result = [self._cache.popleft() for i in range(n)]
        return result

    def peek(self, n=None):
        """
        Preview the next item or `n` items of the iterator.

        The iterator is not advanced when peek is called.

        Args:
            n (int, optional): The number of items to retrieve. Defaults to
                None.

        Returns:
            item or list of items: The next item or `n` items of the iterator.
            If `n` is None, the item itself is returned. If `n` is an int,
            the items will be returned in a list. If `n` is 0, an empty
            list is returned.

            If the iterator is exhausted, `iterpeek.sentinel` is returned,
            or placed as the last item in the returned list:

                >>> p = iterpeek(["a", "b", "c"])
                >>> p.sentinel = "END"
                >>> p.peek()
                'a'
                >>> p.peek(0)
                []
                >>> p.peek(1)
                ['a']
                >>> p.peek(2)
                ['a', 'b']
                >>> p.peek(4)
                ['a', 'b', 'c', 'END']

        Note:
            Will never raise :exc:`StopIteration`.

        """
        self._fillcache(n)
        if n is None:
            result = self._cache[0]
        else:
            result = [self._cache[i] for i in range(n)]
        return result


# Backwards compatibility
peek_iter = iterpeek


class itermod(iterpeek):
    """
    An iterator object that supports modifying items as they are returned.

    >>> a = ["     A list    ",
    ...      "   of strings  ",
    ...      "      with     ",
    ...      "      extra    ",
    ...      "   whitespace. "]
    >>> modifier = lambda s: s.strip().replace('with', 'without')
    >>> for s in itermod(a, modifier=modifier):
    ...   print('"%s"' % s)
    "A list"
    "of strings"
    "without"
    "extra"
    "whitespace."

    Args:
        o (iterable or callable): `o` is interpreted very differently
            depending on the presence of `sentinel`.

            If `sentinel` is not given, then `o` must be a collection object
            which supports either the iteration protocol or the sequence
            protocol.

            If `sentinel` is given, then `o` must be a callable object.

        sentinel (any value, optional): If given, the iterator will call `o`
            with no arguments for each call to its `next` method; if the value
            returned is equal to `sentinel`, :exc:`StopIteration` will be
            raised, otherwise the value will be returned.

        modifier (callable, optional): The function that will be used to
            modify each item returned by the iterator. `modifier` should take
            a single argument and return a single value. Defaults
            to ``lambda x: x``.

            If `sentinel` is not given, `modifier` must be passed as a keyword
            argument.

    Attributes:
        modifier (callable): `modifier` is called with each item in `o` as it
            is iterated. The return value of `modifier` is returned in lieu of
            the item.

            Values returned by `peek` as well as `next` are affected by
            `modifier`. However, `itermod.sentinel` is never passed through
            `modifier`; it will always be returned from `peek` unmodified.

    """

    def __init__(self, *args, **kwargs):
        """__init__(o, sentinel=None, modifier=lambda x: x)"""
        if "modifier" in kwargs:
            self.modifier = kwargs["modifier"]
        elif len(args) > 2:
            self.modifier = args[2]
            args = args[:2]
        else:
            self.modifier = lambda x: x
        if not six.callable(self.modifier):
            raise TypeError("itermod(o, modifier): modifier must be callable")
        super(itermod, self).__init__(*args)

    def _fillcache(self, n):
        """
        Cache `n` modified items. If `n` is 0 or None, 1 item is cached.

        Each item returned by the iterator is passed through the
        `itermod.modified` function before being cached.

        """
        if not n:
            n = 1
        try:
            while len(self._cache) < n:
                self._cache.append(self.modifier(next(self._iterable)))
        except StopIteration:
            while len(self._cache) < n:
                self._cache.append(self.sentinel)


# Backwards compatibility
modify_iter = itermod



# -*- coding: utf-8 -*-
# Copyright (c) 2018 the Pockets team, see AUTHORS.
# Licensed under the BSD License, see LICENSE for details.

"""A pocket full of useful datetime tools!"""

from __future__ import absolute_import, print_function

import sys
from datetime import datetime, timedelta


__all__ = [
    "ceil_datetime",
    "floor_datetime",
    "round_datetime",
    "timedelta_total_seconds",
]


if sys.version_info < (2, 7):

    def timedelta_total_seconds(td):
        """
        Python 2.6 replacement function for timedelta.total_seconds().

        Args:
            td (datetime.timedelta): A `datetime.timedelta` instance.

        Returns:
            float: The total number of seconds plus the fractional
                number of microseconds in `td`.

        """
        total_seconds = td.seconds + (td.days * 24.0 * 3600.0)
        return (td.microseconds + (total_seconds * 1.0e6)) / 1.0e6


else:

    def timedelta_total_seconds(td):
        """
        Python 2.6 replacement function for timedelta.total_seconds().

        Args:
            td (datetime.timedelta): A `datetime.timedelta` instance.

        Returns:
            float: The total number of seconds plus the fractional
                number of microseconds in `td`.

        """
        return td.total_seconds()


def ceil_datetime(dt, nearest):
    """
    Rounds the given `datetime` up to the nearest `timedelta` increment.

    Note:
        `dt.microsecond` is always set to zero and ignored.

    Args:
        dt (datetime.datetime): The `datetime` instance to ceil.
        nearest (datetime.timedelta): The `timedelta` to use as the increment.

    Returns:
        datetime.datetime: A copy of `dt` ceiled to `nearest`.

    """
    dt = dt.replace(microsecond=0)
    dt_min = datetime.min.replace(tzinfo=dt.tzinfo)
    secs = timedelta_total_seconds(dt_min - dt)
    nearest_secs = timedelta_total_seconds(nearest)
    total_delta_secs = secs % nearest_secs
    delta_days = total_delta_secs // 86400  # (60 * 60 * 24)
    delta_secs = total_delta_secs % 86400.0  # (60 * 60 * 24)
    return dt + timedelta(days=delta_days, seconds=delta_secs)


def floor_datetime(dt, nearest):
    """
    Rounds the given `datetime` down to the nearest `timedelta` increment.

    Note:
        `dt.microsecond` is always set to zero and ignored.

    Args:
        dt (datetime.datetime): The `datetime` instance to floor.
        nearest (datetime.timedelta): The `timedelta` to use as the increment.

    Returns:
        datetime.datetime: A copy of `dt` floored to `nearest`.

    """
    dt = dt.replace(microsecond=0)
    dt_min = datetime.min.replace(tzinfo=dt.tzinfo)
    secs = timedelta_total_seconds(dt - dt_min)
    nearest_secs = timedelta_total_seconds(nearest)
    total_delta_secs = secs % nearest_secs
    delta_days = total_delta_secs // 86400  # (60 * 60 * 24)
    delta_secs = total_delta_secs % 86400.0  # (60 * 60 * 24)
    return dt - timedelta(days=delta_days, seconds=delta_secs)


def round_datetime(dt, nearest):
    """
    Rounds the given `datetime` up/down to the nearest `timedelta` increment.

    Note:
        `dt.microsecond` is always set to zero and ignored.

    Args:
        dt (datetime.datetime): The `datetime` instance to round.
        nearest (datetime.timedelta): The `timedelta` to use as the increment.

    Returns:
        datetime.datetime: A copy of `dt` rounded to `nearest`.

    """
    dt = dt.replace(microsecond=0)
    dt_min = datetime.min.replace(tzinfo=dt.tzinfo)
    secs = timedelta_total_seconds(dt - dt_min)
    nearest_secs = timedelta_total_seconds(nearest)
    rounded_secs = ((secs + (nearest_secs / 2)) // nearest_secs) * nearest_secs
    total_delta_secs = rounded_secs - secs
    delta_days = total_delta_secs // 86400  # (60 * 60 * 24)
    delta_secs = total_delta_secs % 86400.0  # (60 * 60 * 24)
    return dt + timedelta(days=delta_days, seconds=delta_secs)



# -*- coding: utf-8 -*-
# Copyright (c) 2018 the Pockets team, see AUTHORS.
# Licensed under the BSD License, see LICENSE for details.

"""A pocket full of useful string manipulation tools!"""

from __future__ import absolute_import, print_function

import re

import six

from pockets.collections import is_listy, listify


__all__ = [
    "camel",
    "uncamel",
    "fieldify",
    "unfieldify",
    "sluggify",
    "splitcaps",
    "splitify",
    "UnicodeMixin",
]


# Default regular expression flags
if six.PY2:
    RE_FLAGS = re.L | re.M | re.U
else:
    RE_FLAGS = re.M | re.U

RE_NONWORD = re.compile(r"[\W_]+")

RE_SPLITCAPS = re.compile(
    # Clause 1
    r"[A-Z]+[^a-z]*"  # All non-lowercase beginning with a capital letter
    r"(?=[A-Z][^A-Z]*?[a-z]|$)"  # Followed by a capitalized word
    r"|"
    # Clause 2
    r"[A-Z][^A-Z]*?[a-z]+[^A-Z]*" r"|"  # Capitalized word
    # Clause 3
    r"[^A-Z]+",  # All non-uppercase
    RE_FLAGS,
)

RE_UNCAMEL = re.compile(
    r"("  # The whole expression is in a single group
    # Clause 1
    r"(?<=[^\sA-Z])"  # Preceded by neither a space nor a capital letter
    r"[A-Z]+[^a-z\s]*"  # All non-lowercase beginning with a capital letter
    r"(?=[A-Z][^A-Z\s]*?[a-z]|\s|$)"  # Followed by a capitalized word
    r"|"
    # Clause 2
    r"(?<=[^\s])"  # Preceded by a character that is not a space
    r"[A-Z][^A-Z\s]*?[a-z]+[^A-Z\s]*"  # Capitalized word
    r")",
    RE_FLAGS,
)

RE_WHITESPACE_GROUP = re.compile(r"(\s+)", RE_FLAGS)


def camel(
    s, sep="_", lower_initial=False, upper_segments=None, preserve_upper=False
):
    """
    Convert underscore_separated string (aka snake_case) to CamelCase.

    Works on full sentences as well as individual words:

    >>> camel("hello_world!")
    'HelloWorld!'
    >>> camel("Totally works as_expected, even_with_whitespace!")
    'Totally Works AsExpected, EvenWithWhitespace!'

    Args:
        sep (string, optional): Delineates segments of `s` that will be
            CamelCased. Defaults to an underscore "_".

            For example, if you want to CamelCase a dash separated word:

            >>> camel("xml-http-request", sep="-")
            'XmlHttpRequest'

        lower_initial (bool, int, or list, optional): If True, the initial
            character of each camelCased word will be lowercase. If False, the
            initial character of each CamelCased word will be uppercase.
            Defaults to False:

            >>> camel("http_request http_response")
            'HttpRequest HttpResponse'
            >>> camel("http_request http_response", lower_initial=True)
            'httpRequest httpResponse'

            Optionally, `lower_initial` can be an int or a list of ints,
            indicating which individual segments of each CamelCased word
            should start with a lowercase. Supports negative numbers to index
            segments from the right:

            >>> camel("xml_http_request", lower_initial=0)
            'xmlHttpRequest'
            >>> camel("xml_http_request", lower_initial=-1)
            'XmlHttprequest'
            >>> camel("xml_http_request", lower_initial=[0, 1])
            'xmlhttpRequest'

        upper_segments (int or list, optional): Indicates which segments of
           CamelCased words should be fully uppercased, instead of just
           capitalizing the first letter.

           Can be an int, indicating a single segment, or a list of ints,
           indicating multiple segments. Supports negative numbers to index
           segments from the right.

           `upper_segments` is helpful when dealing with acronyms:

            >>> camel("tcp_socket_id", upper_segments=0)
            'TCPSocketId'
            >>> camel("tcp_socket_id", upper_segments=[0, -1])
            'TCPSocketID'
            >>> camel("tcp_socket_id", upper_segments=[0, -1], lower_initial=1)
            'TCPsocketID'

        preserve_upper (bool): If True, existing uppercase characters will
            not be automatically lowercased. Defaults to False.

            >>> camel("xml_HTTP_reQuest")
            'XmlHttpRequest'
            >>> camel("xml_HTTP_reQuest", preserve_upper=True)
            'XmlHTTPReQuest'

    Returns:
        str: CamelCased version of `s`.

    """
    if isinstance(lower_initial, bool):
        lower_initial = [0] if lower_initial else []
    else:
        lower_initial = listify(lower_initial)
    upper_segments = listify(upper_segments)
    result = []
    for word in RE_WHITESPACE_GROUP.split(s):
        segments = [segment for segment in word.split(sep) if segment]
        count = len(segments)
        for i, segment in enumerate(segments):
            upper = i in upper_segments or (i - count) in upper_segments
            lower = i in lower_initial or (i - count) in lower_initial
            if upper and lower:
                if preserve_upper:
                    segment = segment[0] + segment[1:].upper()
                else:
                    segment = segment[0].lower() + segment[1:].upper()
            elif upper:
                segment = segment.upper()
            elif lower:
                if not preserve_upper:
                    segment = segment.lower()
            elif preserve_upper:
                segment = segment[0].upper() + segment[1:]
            else:
                segment = segment[0].upper() + segment[1:].lower()
            result.append(segment)

    return "".join(result)


def uncamel(s, sep="_"):
    """
    Convert CamelCase string to underscore_separated (aka snake_case).

    A CamelCase word is considered to be any uppercase letter followed by zero
    or more lowercase letters. Contiguous groups of uppercase letters – like
    you would find in an acronym – are also considered part of a single word:

    >>> uncamel("Request")
    'request'
    >>> uncamel("HTTP")
    'http'
    >>> uncamel("HTTPRequest")
    'http_request'
    >>> uncamel("xmlHTTPRequest")
    'xml_http_request'

    Works on full sentences as well as individual words:

    >>> uncamel("HelloWorld!")
    'hello_world!'
    >>> uncamel("Totally works AsExpected, EvenWithWhitespace!")
    'totally works as_expected, even_with_whitespace!'

    Args:
        sep (str, optional): String used to separate CamelCase words. Defaults
            to an underscore "_".

            For example, if you want dash separated words:

            >>> uncamel("XmlHttpRequest", sep="-")
            'xml-http-request'

    Returns:
        str: uncamel_cased version of `s`.

    """
    return RE_UNCAMEL.sub(r"{0}\1".format(sep), s).lower()


def fieldify(s, sep="_"):
    """
    Convert a string into a valid "field-like" variable name.

    Converts `s` from camel case to underscores, and replaces all spaces and
    non-word characters with `sep`:

    >>> fieldify('The XmlHTTPRequest Contained, "DATA..."')
    'the_xml_http_request_contained_data'

    Args:
        s (str): The string to fieldify.

        sep (str): The string to use as a word separator in the returned field.
            Defaults to '_'.

    Returns:
        str: The field version of `s`.

    """
    if not s:
        return ""
    return RE_NONWORD.sub(sep, uncamel(s)).strip(sep)


def unfieldify(s, sep="_"):
    """
    Makes a best effort to reverse the algorithm from `fieldify`.

    Replaces instances of `sep` in `s` with a space and converts the result to
    title case:

    >>> unfieldify('the_xml_http_request_contained_data')
    'The Xml Http Request Contained Data'

    Args:
        s (str): The string to fieldify.

        sep (str): The string to consider a word separator in `s`.
            Defaults to '_'.

    Returns:
        str: The unfieldified version of `s`.

    """
    if not s:
        return ""
    s = s.strip(r"{0} ".format(sep))
    return (" ".join([w for w in s.split(sep) if w])).title()


def sluggify(s, sep="-"):
    """
    Convert a string into a "slug" suitable for use in a URL.

    Converts `s` to lower case, and replaces all spaces and non-word
    characters with `sep`:

    >>> sluggify('The ANGRY Wizard Shouted, "HEY..."')
    'the-angry-wizard-shouted-hey'

    Args:
        s (str): The string to convert into a slug.

        sep (str): The string to use as a word separator in the slug.
            Defaults to '-'.

    Returns:
        str: The sluggify version of `s`.

    """
    if not s:
        return ""
    return RE_NONWORD.sub(sep, s).lower().strip(sep)


def splitcaps(s, pattern=None, maxsplit=None, flags=0):
    """
    Intelligently split a string on capitalized words.

    A capitalized word is considered to be any uppercase letter followed by
    zero or more lowercase letters. Contiguous groups of uppercase letters –
    like you would find in an acronym – are also considered part of a single
    word:

    >>> splitcaps("Request")
    ['Request']
    >>> splitcaps("HTTP")
    ['HTTP']
    >>> splitcaps("HTTPRequest")
    ['HTTP', 'Request']
    >>> splitcaps("HTTP/1.1Request")
    ['HTTP/1.1', 'Request']
    >>> splitcaps("xmlHTTPRequest")
    ['xml', 'HTTP', 'Request']

    If no capitalized words are found in `s`, the whole string is
    returned in a single element list:

    >>> splitcaps("")
    ['']
    >>> splitcaps("lower case words")
    ['lower case words']

    Does not split on whitespace by default. To also split
    on whitespace, pass "\\\\s+" for `pattern`:

    >>> splitcaps("Without whiteSpace pattern")
    ['Without white', 'Space pattern']
    >>> splitcaps("With whiteSpace pattern", pattern=r"\\s+")
    ['With', 'white', 'Space', 'pattern']
    >>> splitcaps("With whiteSpace group", pattern=r"(\\s+)")
    ['With', ' ', 'white', 'Space', ' ', 'group']

    Args:
        s (str): The string to split.

        pattern (str, optional): In addition to splitting on capital letters,
            also split by the occurrences of `pattern`. If capturing
            parentheses are used in `pattern`, then the text of all groups in
            `pattern` are also returned as part of the resulting list.
            Defaults to None.

        maxsplit (int, optional):  If maxsplit is not specified or -1, then
            there is no limit on the number of splits (all possible splits are
            made). If maxsplit is >= 0, at most maxsplit splits occur, and the
            remainder of the string is returned as the final element of the
            list.

        flags (int, optional): Flags to pass to the regular expression created
            using `pattern`. Ignored if `pattern` is not specified. Defaults
            to (re.LOCALE | re.MULTILINE | re.UNICODE).

    Returns:
        list: List of capitalized substrings in `s`.

    """
    if not maxsplit:
        if maxsplit == 0:
            return [s]
        else:
            maxsplit = -1

    if pattern:
        pattern_re = re.compile(pattern, flags or RE_FLAGS)
    else:
        pattern_re = None

    result = []
    post_maxsplit = []
    for m in RE_SPLITCAPS.finditer(s):
        if pattern_re:
            for segment in pattern_re.split(m.group()):
                if segment:
                    if maxsplit > 0 and len(result) >= maxsplit:
                        post_maxsplit.append(segment)
                    else:
                        result.append(segment)
        else:
            result.append(m.group())

        if maxsplit > 0 and len(result) >= maxsplit:
            if m.end() < len(s):
                post_maxsplit.append(s[m.end() :])
            post_maxsplit = "".join(post_maxsplit)
            if post_maxsplit:
                result.append(post_maxsplit)
            break

    return result if len(result) > 0 else [s]


def splitify(value, separator=",", strip=True, include_empty=False):
    """
    Convert a value to a list using a supercharged `split()`.

    If `value` is a string, it is split by `separator`. If `separator` is
    `None` or empty, no attempt to split is made, and `value` is returned as
    the only item in a list.

    If `strip` is `True`, then the split strings will be stripped of
    whitespace. If `strip` is a string, then the split strings will be
    stripped of the given string.

    If `include_empty` is `False`, then empty split strings will not be
    included in the returned list.

    If `value` is `None` an empty list is returned.

    If `value` is already "listy", it is returned as-is.

    If `value` is any other type, it is returned as the only item in a list.

    >>> splitify("first item, second item")
    ['first item', 'second item']
    >>> splitify("first path: second path: :skipped empty path", ":")
    ['first path', 'second path', 'skipped empty path']
    >>> splitify(["already", "split"])
    ['already', 'split']
    >>> splitify(None)
    []
    >>> splitify(1969)
    [1969]
    """
    if is_listy(value):
        return value

    if isinstance(value, str) and separator:
        parts = value.split(separator)
        if strip:
            strip = None if strip is True else strip
            parts = [s.strip(strip) for s in parts]
        return [s for s in parts if include_empty or s]
    return listify(value)


class UnicodeMixin(object):
    """
    Mixin class to define proper __str__/__unicode__ methods in Python 2 or 3.

    Originally found on the `Porting Python 2 Code to Python 3 HOWTO`_.

    .. _Porting Python 2 Code to Python 3 HOWTO:
       https://docs.python.org/3.3/howto/pyporting.html

    """

    if six.PY2:

        def __str__(self):
            return self.__unicode__().encode("utf8")

    else:

        def __str__(self):
            return self.__unicode__()



# -*- coding: utf-8 -*-
# Copyright (c) 2018 the Pockets team, see AUTHORS.
# Licensed under the BSD License, see LICENSE for details.

"""
*Let me check my pockets...*

Functions available in the `pockets.*` submodules are also imported to the base
package for easy access, so::

    from pockets import camel, iterpeek, resolve

works just as well as::

    from pockets.inspect import resolve
    from pockets.iterators import iterpeek
    from pockets.string import camel

"""

from __future__ import absolute_import, print_function

import sys

from pockets.inspect import hoist_submodules


hoist_submodules(sys.modules[__name__])



# -*- coding: utf-8 -*-
# Copyright (c) 2018 the Pockets team, see AUTHORS.
# Licensed under the BSD License, see LICENSE for details.

"""
An easy to import AutoLogger instance!

>>> import logging, sys
>>> logging.basicConfig(format="%(name)s: %(message)s", stream=sys.stdout)
>>> from pockets.autolog import log
>>> log.error("Always log from the correct module.Class!") # doctest: +SKIP
pockets.autolog: Always log from the correct module.Class!

See Also:
    `pockets.logger.AutoLogger`
"""

from __future__ import absolute_import, print_function

from pockets.logging import AutoLogger


__all__ = ["log"]


log = AutoLogger()



# -*- coding: utf-8 -*-
# Copyright (c) 2018 the Pockets team, see AUTHORS.
# Licensed under the BSD License, see LICENSE for details.

"""
A pocket full of useful logging tools!

The `pockets.logging` module adds a `logging.TRACE` level and a
`logging.Logger.trace` method, so messages can be logged at a lower priority
level than `logging.DEBUG`.
"""

from __future__ import absolute_import, print_function

import inspect
import logging
import logging.config
import sys
from functools import wraps

from six import text_type as unicode_type


__all__ = [
    "log_exceptions",
    "AutoLogger",
    "EagerFormattingAdapter",
    "IndentMultilineLogFormatter",
]


TRACE = 5
logging.addLevelName(TRACE, "TRACE")
logging.TRACE = TRACE


def trace(self, message, *args, **kwargs):
    # Yes, _log() takes '*args' as 'args'.
    self._log(TRACE, message, args, **kwargs)


logging.Logger.trace = trace


def log_exceptions(fn):
    """
    Decorator that wraps a function and logs any raised exceptions.

    The exception will still be raised after being logged. Also logs the
    arguments to every call at the trace level.
    """
    from pockets.autolog import log

    @wraps(fn)
    def wrapper(*args, **kwargs):
        try:
            a = [str(x)[:255] for x in args]
            kw = dict([(k[:255], str(v)[:255]) for k, v in kwargs.items()])
            log.trace("Calling %s.%s %r %r", fn.__module__, fn.__name__, a, kw)
            return fn(*args, **kwargs)
        except Exception as e:
            log.error("Error calling function %s: %s" % (fn.__name__, e))
            log.exception(e)
            raise

    return wrapper


class AutoLogger(object):
    """
    A logger proxy object with all of the methods and attributes of `Logger`.

    When an attribute (e.g., "debug") is requested, inspects the stack for the
    calling module's name, and passes that name to `logging.getLogger`.

    What this means is that you can instantiate an `AutoLogger` anywhere, and
    when you call it, the log entry shows the module where you called it, not
    where it was created.

    `AutoLogger` also inspects the local variables where it is called, looking
    for `self`. If `self` exists, its classname is added to the module name.

    Args:
        adapter_class (LoggerAdapter): optional `LoggerAdapter` class to use.
        adapter_args (list): optional args to use when instantiating an
            instance of `adapter_class`.
        adapter_kwargs (dict): optional kwargs to use when instantiating an
            instance of `adapter_class`.
    """

    def __init__(
        self, adapter_class=None, adapter_args=None, adapter_kwargs=None
    ):
        if adapter_args is None:
            adapter_args = []
        if adapter_kwargs is None:
            adapter_kwargs = {}

        self.adapter_class = adapter_class
        self.adapter_args = adapter_args
        self.adapter_kwargs = adapter_kwargs

    def __getattr__(self, name):
        f_locals = inspect.currentframe().f_back.f_locals
        if "self" in f_locals and f_locals["self"] is not None:
            other = f_locals["self"]
            caller_name = "%s.%s" % (
                other.__class__.__module__,
                other.__class__.__name__,
            )
        else:
            caller_name = inspect.currentframe().f_back.f_globals["__name__"]
        logger = logging.getLogger(caller_name)

        if self.adapter_class:
            logger = self.adapter_class(
                logger, *self.adapter_args, **self.adapter_kwargs
            )

        return getattr(logger, name)


class EagerFormattingAdapter(logging.LoggerAdapter):
    """
    A `LoggerAdapter` that immediately interpolates message arguments if the
    appropriate loglevel is set.

    This is useful because many log handlers generate log output on a separate
    thread, and the value of the log arguments may have changed by the time
    the handler interpolates them. This can lead to confusion when debugging
    difficult bugs, as the log output will not reflect what was actually
    happening when the log message was originally generated.

    For performance reasons, the interpolation ONLY happens if the appropriate
    loglevel is set. This prevents unnecessary string formatting on log
    messages that will just be thrown out anyway.

    Args:
        logger (Logger): The underlying Logger instance to use.
        extra (dict): Extra args, ignored by this implementation.
    """

    def __init__(self, logger, extra=None):
        """
        Initialize the adapter with a logger and a dict-like object which
        provides contextual information. This constructor signature allows
        easy stacking of LoggerAdapters, if so desired.

        You can effectively pass keyword arguments as shown in the
        following example::

            adapter = LoggerAdapter(someLogger, dict(p1=v1, p2="v2"))

        """
        self.logger = logger
        self.extra = extra

    def _eagerFormat(self, msg, level, args):
        """
        Eagerly apply log formatting if the appropriate level is enabled.

        Otherwise we just drop the log message (and return a string indicating
        that it was suppreseed).
        """
        if not hasattr(self, "isEnabledFor") or self.isEnabledFor(level):
            # Do the string formatting immediately.
            if args:
                return self._getUnterpolatedMessage(msg, args)
            else:
                return msg
        else:
            # Otherwise, just drop the message completely to avoid anything
            # going wrong in the future.  This text shoudl clue one in to
            # what's going on in the bizarre edge case where this ever does
            # show up.
            return "(log message suppressed due to insufficient log level)"

    def _getUnterpolatedMessage(self, msg, args):
        """
        Returns the formatted string, will first attempt str.format and will
        fallback to msg % args as it was originally.

        This is lifted almost wholesale from logging_unterpolation.
        """
        original_msg = msg

        try:
            msg = msg.format(*args)
        except UnicodeEncodeError:
            # This is most likely due to formatting a non-ascii string argument
            # into a bytestring, which the %-operator automatically handles
            # by casting the left side (the "msg" variable) in this context
            # to unicode. So we'll do that here
            #
            # Handle the attempt to print utf-8 encoded data, similar to
            # %-interpolation's handling of unicode formatting non-ascii
            # strings
            msg = unicode_type(msg).format(*args)

        except ValueError:
            # From PEP-3101, value errors are of the type raised by the format
            # method itself, so see if we should fall back to original
            # formatting since there was an issue
            if "%" in msg:
                msg = msg % args
            else:
                # We should NOT fall back, since there's no possible string
                # interpolation happening and we want a meaningful error
                # message
                raise

        if msg == original_msg and "%" in msg:
            # There must have been no string formatting methods used, given
            # the presence of args without a change in the msg
            if len(args) == 1 and isinstance(args[0], dict):
                # Handles cases like:
                # logging.debug("a %(a)d b %(b)s", {'a':1, 'b':2})
                msg = msg % args[0]
            else:
                # Fall back to original formatting
                msg = msg % args

        return msg

    def trace(self, msg, *args, **kwargs):
        """
        Delegate a trace call to the underlying logger, after adding
        contextual information from this adapter instance.
        """
        self.log(logging.TRACE, msg, *args, **kwargs)

    def debug(self, msg, *args, **kwargs):
        """
        Delegate a debug call to the underlying logger, after adding
        contextual information from this adapter instance.
        """
        self.log(logging.DEBUG, msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        """
        Delegate an info call to the underlying logger, after adding
        contextual information from this adapter instance.
        """
        self.log(logging.INFO, msg, *args, **kwargs)

    def warn(self, msg, *args, **kwargs):
        """
        Delegate a warning call to the underlying logger, after adding
        contextual information from this adapter instance.
        """
        self.log(logging.WARN, msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        """
        Delegate a warning call to the underlying logger, after adding
        contextual information from this adapter instance.
        """
        self.log(logging.WARNING, msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        """
        Delegate an error call to the underlying logger, after adding
        contextual information from this adapter instance.
        """
        self.log(logging.ERROR, msg, *args, **kwargs)

    def exception(self, msg, *args, **kwargs):
        """
        Delegate an exception call to the underlying logger, after adding
        contextual information from this adapter instance.
        """
        kwargs["exc_info"] = 1
        self.log(logging.ERROR, msg, *args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        """
        Delegate a critical call to the underlying logger, after adding
        contextual information from this adapter instance.
        """
        self.log(logging.CRITICAL, msg, *args, **kwargs)

    def fatal(self, msg, *args, **kwargs):
        """
        Delegate a fatal call to the underlying logger, after adding
        contextual information from this adapter instance.
        """
        self.log(logging.FATAL, msg, *args, **kwargs)

    def log(self, level, msg, *args, **kwargs):
        """
        Delegate a log call to the underlying logger, after adding
        contextual information from this adapter instance.
        """
        msg, kwargs = self.process(msg, kwargs)
        # We explicitly do not pass the args into the log method here, since
        # they should be "used up" by the eagerFormat method.
        self.logger.log(level, self._eagerFormat(msg, level, args), **kwargs)


class IndentMultilineLogFormatter(logging.Formatter):
    """
    Formatter which indents messages that are split across multiple lines.

    Indents all lines that start with a newline so they are easier for
    external log programs to parse.
    """

    def format(self, record):
        """
        Formats the given `LogRecord` by indenting all newlines.

        Args:
            record (LogRecord): The `LogRecord` to format.

        Returns:
            str: The formatted message with all newlines indented.
        """
        if sys.version_info < (2, 7):
            s = logging.Formatter.format(self, record)
        else:
            s = super(IndentMultilineLogFormatter, self).format(record)
        return s.rstrip("\n").replace("\n", "\n  ")
